"""
This module contains several function to either:
1. Help server.py and client.py under-the-hood
2. Help the user by providing them with some built-in functions

Generally, functions starting with an underscore (_) will be
under-the-hood, while the rest are user functions

====================================
Copyright SSS_Says_Snek, 2021-present
====================================
"""

from __future__ import annotations

import re
import socket
from typing import Union, Optional


# __all__ = [
#     'make_header', 'receive_message',
#     'get_local_ip', 'ipstr_to_tup',
#     'iptup_to_str'
# ]


# Some custom exceptions
class ClientException(Exception):
    pass


class ServerException(Exception):
    pass


class NoMessageException(Exception):
    pass


class ServerNotRunning(Exception):
    pass


class ClientDisconnected(Exception):
    pass


# Custom warnings
class NoHeaderWarning(Warning):
    pass


def make_header(header_msg: Union[str, bytes],
                header_len: int, encode=True) -> Union[str, bytes]:
    """
    Makes a header of ``header_msg``, with a maximum
    header length of ``header_len``

    :param header_msg: A string OR bytes-like object, representing
        the data to make a header from
    :type header_msg: Union[str, bytes]
    :param header_len: An integer, specifying
        the actual header length (will be padded)
    :type header_len: int
    :param encode: A boolean, specifying the
    :return: The constructed header, padded to ``header_len``
        bytes
    :rtype: Union[str, bytes]
    """
    len_msg = len(header_msg)
    constructed_header = f"{len_msg}{' ' * (header_len - len(str(len_msg)))}"
    if encode:
        return constructed_header.encode()
    return constructed_header


def receive_message(connection, header_len) -> dict:
    """
    Receives a message from a server or client.

    :param connection: The socket to listen to for messages.
        **MUST BE A SOCKET**
    :type connection: socket.socket
    :param header_len: The length of the header, so that
        it can successfully retrieve data without loss/gain of data
    :type header_len: int
    :return: A dictionary, with two key-value pairs;
        The first key-value pair refers to the header,
        while the second one refers to the actual data
    :rtype: dict["header": bytes, "data": bytes]
    """
    try:
        header_msg = connection.recv(header_len)

        if header_msg:
            msg_len = int(header_msg)
            data = connection.recv(msg_len)

            return {"header": header_msg, "data": data}
        raise NoMessageException(
            "No header received, aborting..."
        )
    except ConnectionResetError:
        pass


def _removeprefix(
        string: Union[str, bytes],
        prefix: Union[str, bytes],
) -> Union[str, bytes]:
    """A backwards-compatible alternative of str.removeprefix"""
    if string.startswith(prefix):
        return string[len(prefix):]
    else:
        return string[:]


def _dict_tupkey_lookup(multikey, _dict, idx_to_match=None):
    """
    Returns the value of the dict looked up,
    given a key that is part of a key-tuple
    """
    for key, value in _dict.items():
        if idx_to_match is None:
            if multikey in key:
                yield value
        elif isinstance(idx_to_match, int):
            if multikey == key[idx_to_match]:
                yield value


def _dict_tupkey_lookup_key(multikey, _dict, idx_to_match=None):
    """
    Returns the key of the dict looked up,
    given a key that is part of a key-tuple
    """
    for key in _dict.keys():
        if idx_to_match is None:
            if multikey in key:
                yield key
        elif isinstance(idx_to_match, int):
            if multikey == key[idx_to_match]:
                yield key


def _type_cast_server(
        type_cast,
        content_to_typecast: bytes,
        func_dict: dict
):
    """
    Basis for type casting on the server
    If testing, replace `func_dict` with a dummy one
    Currently NOT guarenteed to return, please remember to change this API
    """
    if type_cast == str:
        try:
            typecasted_content = content_to_typecast.decode()
            return typecasted_content  # Remember to change this, but I"m lazy rn
        except UnicodeDecodeError as e:
            raise TypeError(
                f"Type casting from bytes to string failed for function "
                f"\"{func_dict['name']}\"\n{str(e)}"
            )
    elif type_cast == int:
        try:
            typecasted_content = int(content_to_typecast)
            return typecasted_content  # Remember to change this, but I"m lazy rn
        except ValueError as e:
            raise TypeError(
                f"Type casting from bytes to int failed for function "
                f"\"{func_dict['name']}\":\n           {e}"
            ) from ValueError
    elif type_cast == float:
        try:
            typecasted_content = float(content_to_typecast)
            return typecasted_content  # Remember to change this, but I"m lazy rn
        except ValueError as e:
            raise TypeError(
                f"Type casting from bytes to float failed for function "
                f"\"{func_dict['name']}\":\n           {e}"
            ) from ValueError
    # elif type_cast == list:
    #     try:
    #         _dict = json.loads(content_to_typecast)
    #         typecasted_content = list(_dict.values())
    #         return typecasted_content  # Remember to change this, but I"m lazy rn
    #     except json.decoder.JSONDecodeError as e:
    #         raise TypeError(
    #             f"Type casting from bytes to list"
    #         )


def get_local_ip():
    """
    Gets the local IP of your device, with sockets

    :return: A string containing the IP address, in the
        format "ip:port"
    :rtype: str
    """
    return socket.gethostbyname(socket.gethostname())


def input_server_config(
        ip_prompt: str = "Enter the IP of where to host the server: ",
        port_prompt: str = "Enter the Port of where to host the server: "
) -> tuple[str, int]:
    """
    Provides a built-in way to obtain the IP and port of where the server
    should be hosted, through :func:`input()`

    :param ip_prompt: A string, specifying the prompt to show when
        asking for IP.

        Default is "Enter the IP of where to host the server: "
    :type ip_prompt: str, optional
    :param port_prompt: A string, specifying the prompt to show when
        asking for Port

        Default is "Enter the Port of where to host the server: "
    :type port_prompt: str, optional
    :return: A two-element tuple, consisting of IP and Port
    :rtype: tuple[str, int]
    """

    # Flags
    ip_range_check = True

    ip = input(ip_prompt)

    if re.search("^((\d?){3}\.){3}(\d\d?\d?)[ ]*$", ip):
        # IP conformity regex
        split_ip = list(map(int, ip.split('.')))
        split_ip = [i > 255 for i in split_ip]

        if any(split_ip):
            ip_range_check = True

    while (ip == '' or not re.search(
            "^((\d?){3}\.){3}(\d\d?\d?)[ ]*$",
            ip
    )) or ip_range_check:
        # If IP not conform to regex, accept input until it
        # is compliant
        ip = input(f"\033[91mE: Invalid IP\033[0m\n{ip_prompt}")
        if re.search("^((\d?){3}\.){3}(\d\d?\d?)[ ]*$", ip):
            split_ip = list(map(int, ip.split('.')))
            split_ip = [i > 255 for i in split_ip]

            if not any(split_ip):
                ip_range_check = False

    port = input(port_prompt)

    while (
            (port == '' or not port.isdigit()) or
            (port.isdigit() and int(port) > 65535)
    ):
        # If port is > 65535, or not numerical, repeat until it is
        port = input(f"\033[91mE: Invalid Port\033[0m\n{port_prompt}")
    port = int(port)

    # Returns
    return ip, port


def input_client_config(
        ip_prompt: str = "Enter the IP of the server: ",
        port_prompt: str = "Enter the Port of the server: ",
        name_prompt: Union[str, None] = "Enter name to connect as: ",
        group_prompt: Union[str, None] = "Enter group to connect to: "
) -> tuple[Union[str, int], ...]:
    """
    Provides a built-in way to obtain the IP and port of the configuration
    of the server to connect to

    :param ip_prompt: A string, specifying the prompt to show when
        asking for IP.

        Default is "Enter the IP of the server: "
    :type ip_prompt: str, optional
    :param port_prompt: A string, specifying the prompt to show when
        asking for Port

        Default is "Enter the Port of the server: "
    :type port_prompt: str, optional
    :param name_prompt: A string, specifying the prompt to show when
        asking for Client Name

        Default is "Enter name to connect as: " (Pass in None for no input)
    :type name_prompt: Union[str, None], optional
    :param group_prompt: A string, specifying the prompt to show when
        askign for Client Group

        Default is "Enter group to connect to: " (Pass in None for no input)
    :type group_prompt: Union[str, None], optional
    :return: A tuple containing the config options of the server
    :rtype: tuple[str, int, Optional[str], Optional[int]]
    """
    # Grabs IP and Port from previously defined function
    ip, port = input_server_config(
        ip_prompt, port_prompt
    )

    # Flags
    name = None
    group = None

    # If names are enabled
    if name_prompt is not None:
        name = input(name_prompt)

        # Repeat until name has an input
        while name == '':
            name = input(name_prompt)

    # If groups are enabled
    if group_prompt is not None:
        group = input(group_prompt)

        # Repeat until group has an input
        while group == '':
            group = input(group_prompt)

    # Return list
    ret = [ip, port]

    if name is not None:
        ret.append(name)
    if group is not None:
        ret.append(group)

    # Return
    return tuple(ret)


def ipstr_to_tup(
        formatted_ip: str
) -> tuple[Union[str, int], ...]:
    """
    Converts a string IP address into a tuple equivalent

    :param formatted_ip: A string, representing the IP address.

        Must be in the format "ip:port"
    :type formatted_ip: str

    :return: A tuple, with IP address as the first element, and
        an INTEGER port as the second element
    :rtype: tuple[str, int]
    """
    ip_split = formatted_ip.split(':')
    recon_ip_split = [
        str(ip_split[0]),
        int(ip_split[1])
    ]
    return tuple(recon_ip_split)  # Convert list to tuple


def iptup_to_str(
        formatted_tuple: tuple[str, int]
) -> str:
    """
    Converts a tuple IP address into a string equivalent

    This function is like the opposite of ``ipstr_to_tup``

    :param formatted_tuple: A two-element tuple, containing the IP address and the port.
        Must be in the format (ip: str, port: int)
    :type formatted_tuple: tuple[str, int]

    :return: A string, with the format "ip:port"
    :rtype: str
    """
    return f"{formatted_tuple[0]}:{formatted_tuple[1]}"



import os
import sys
import pathlib

PATH = pathlib.Path(__file__)

# Import stuff, as this situation's a bit wonky
# This is not needed in usual hisock code. However, in this odd case,
# imports need to be configured first like this
sys.path.append(str(PATH.parent.parent.parent))  # Not needed usually; used to run examples

from hisock import (
    connect
)


# Data class, as to prevent using those dirty globals
class Data:
    def __init__(self):
        self.tictactoe_opponent = None  # Tic Tac Toe opponent name
        self.board = [' ' for _ in range(9)]  # Tic Tac Toe board
        self.board_layout = "{} |{} |{} \n--+--+--\n{} |{} |{} \n--+--+--\n{} |{} |{} \n"
        self.letter = ''


def clear():
    """Well, no pycharm support, so boo hoo"""
    if os.name == "nt":
        os.system("cls")
    elif os.name == "posix":
        os.system("clear")


def run():
    # IP, Port, and Name establishment
    ip_input = input("Enter IP of where server is hosted: ")

    port_input = input("Enter port of server: ")
    while port_input == '':
        port_input = input("Enter port of server: ")
    port_input = int(port_input)

    name = input("Enter name (MAX LENGTH: 32): ")
    while name == '' or len(name) > 32:
        name = input("Enter name (MAX LENGTH: 32): ")

    print(f"Connecting to server at {ip_input}:{port_input}...", end=" ")

    server = connect((ip_input, port_input), name)

    print("SUCCESS!")
    clear()
    print("Waiting for pairing... ")

    data = Data()

    # Hisock message receive decorators
    @server.on("game_start")
    def game_started(opponent: str):
        print(f"Pairing succeeded! Opponent: \"{opponent}\"")
        data.tictactoe_opponent = opponent
        goes_first = server.recv_raw()

        if goes_first == b"First":
            # Receives message from server; goes first, therefore is X
            data.letter = "X"
            print("You are X")
            print(data.board_layout.format(*data.board))
            move_input = input("Enter your move (1-9, left to right, top to bottom): ")
            while (
                    move_input == '' or not move_input.isdigit() or
                    (move_input.isdigit() and not 1 <= int(move_input) <= 9)
            ):
                move_input = input("Enter your move (1-9, left to right, top to bottom): ")
            move_input = int(move_input)
            data.board[move_input - 1] = "X"

            # Refreshes the screen
            clear()
            print("You are X")
            print(data.board_layout.format(*data.board))

            # Sends to server the index of the move
            server.send("player_turn", str(move_input - 1).encode())
        else:
            # Goes last, therefore is O
            data.letter = "O"
            print("You are O")
            print(data.board_layout.format(*data.board))

    @server.on("player_turn")
    def player_turn(move: int):
        # GETS CALLED EVERY TURN AFTER FIRST DETERMINATION

        data.board[move] = "X" if data.letter == "O" else "O"

        # Refreshes the screen
        clear()
        print(f"You are {data.letter}")
        print(data.board_layout.format(*data.board))

        move_input = input("Enter your move (1-9, left to right, top to bottom): ")
        while (
                move_input == '' or not move_input.isdigit() or
                (move_input.isdigit() and not 1 <= int(move_input) <= 9) or
                (move_input.isdigit() and data.board[int(move_input) - 1] != " ")
        ):
            move_input = input("Enter your move (1-9, left to right, top to bottom): ")
        move_input = int(move_input)
        data.board[move_input - 1] = data.letter

        # Refreshes the screen after input
        clear()
        print(f"You are {data.letter}")
        print(data.board_layout.format(*data.board))

        server.send("player_turn", str(move_input - 1).encode())

    @server.on("win")
    def player_win(missing_move: int):
        # Player won

        if data.board[missing_move] == " ":
            data.board[missing_move] = "X" if data.letter == "O" else "O"

        # Refreshes the screen
        clear()
        print("You win!")
        print(data.board_layout.format(*data.board))

        sys.exit()

    @server.on("lose")
    def player_lose(missing_move: int):
        # Player lost

        if data.board[missing_move] == " ":
            data.board[missing_move] = "X" if data.letter == "O" else "O"

        # Refreshes the screen
        clear()
        print("You lose!")
        print(data.board_layout.format(*data.board))

        sys.exit()

    @server.on("tie")
    def player_tie(missing_move: int):
        # Players tied

        if data.board[missing_move] == " ":
            data.board[missing_move] = "X" if data.letter == "O" else "O"

        # Refreshes the screen
        clear()
        print("You tied!")
        print(data.board_layout.format(*data.board))

        sys.exit()

    @server.on("opp_disc")
    def opponent_leave(opp_name: str):
        clear()
        print(f"Opponent \"{opp_name}\"disconnected; disconnecting from server...")
        raise SystemExit

    while True:
        # Updates server
        server.update()


if __name__ == "__main__":
    run()



# Imports

import pathlib
import random
import statistics
import sys

# Import stuff, as this situation's a bit wonky
# This is not needed in usual hisock code. However, in this odd case,
# imports need to be configured first like this

sys.path.append(str(pathlib.Path(__file__).parent.parent.parent))  # Not needed usually; used to run examples

from hisock import start_server, utils


# Data class, as to prevent using those dirty globals
class Data:
    def __init__(self):
        self.boards = []


# IP and Port establishment

def run():
    if len(sys.argv) - 1 != 1:
        ip_input = input("Enter IP of where the server should be hosted "
                         "(Leave blank for current IP): ")
        if ip_input == '':
            ip_input = utils.get_local_ip()

        port_input = input("Enter port of server: ")
        while port_input == '':
            port_input = input("Enter port of server: ")

        port_input = int(port_input)
    else:
        ip_input = utils.get_local_ip()
        port_input = 6969

    print(f"Starting server at {ip_input}:{port_input}...", end=" ")
    s = start_server((ip_input, port_input))
    print("SUCCESS!")

    # Tic Tac Toe variable initialization
    paired_clients = []
    paired_clients_ip = []
    x_or_o = []

    data = Data()

    # Hisock message receive decorators
    @s.on("join")
    def client_conn(client_info):
        print(f"Client \"{client_info['name']}\" connected")

        # Gathers last pair/user from the list
        try:
            last_pair = paired_clients[-1]
            last_ip_pair = paired_clients_ip[-1]
        except IndexError:
            last_pair = ("Place", "Holder")  # To conform to the 2-1 last pair searchup
            last_ip_pair = ("Place", "Holder")

        if len(last_ip_pair) == 1:  # Pair client to currently available one
            print(f"    - Paired with \"{last_pair[0]}\"")
            paired_clients[-1].append(client_info['name'])
            paired_clients_ip[-1].append(client_info['ip'])

            goes_first = random.random()

            s.send_client(client_info['ip'], "game_start", last_pair[0].encode())
            s.send_client(last_ip_pair[0], "game_start", client_info['name'].encode())

            if goes_first < 0.5:
                # Connected client goes first
                s.send_client_raw(client_info['ip'], b"First")
                s.send_client_raw(last_ip_pair[0], b"Last")
                x_or_o.append(["O", "X"])
            else:
                # Unpaired client goes first
                s.send_client_raw(last_ip_pair[0], b"First")
                s.send_client_raw(client_info['ip'], b"Last")
                x_or_o.append(["X", "O"])

        elif len(last_pair) == 2:  # Available client
            # Appended to variables - no other action taken
            print("    - Currently available (no clients to pair)")
            paired_clients.append([client_info['name']])
            paired_clients_ip.append([client_info['ip']])

    @s.on("leave")
    def player_leave(clt_info):
        print(paired_clients_ip)
        print(f"Client \"{clt_info['name']}\" disconnected")
        for ip_pair in paired_clients_ip[:]:
            for ip in ip_pair:
                if ip == clt_info['ip']:
                    if len(ip_pair) == 1:
                        paired_clients_ip.remove(ip_pair)
                    else:
                        # Uhh I don't know what to do RN
                        cp_ip_pair = ip_pair[:]
                        cp_ip_pair.remove(ip)
                        other_ip = cp_ip_pair[0]
                        s.send_client(other_ip, "opp_disc", clt_info['name'].encode())
                        paired_clients_ip.remove(ip_pair)
        print(paired_clients_ip)

    @s.on("player_turn")
    def player_turn(clt_data, move: int):
        opponent_ip = None  # Suppressing Pycharm warnings
        opponent_letter = None
        idx = None

        # Locates index of player pair
        for x, ip_pair in enumerate(paired_clients_ip):
            for y, ip in enumerate(ip_pair):
                if ip == clt_data['ip']:
                    opponent_ip = paired_clients_ip[x][abs(y - 1)]
                    opponent_letter = x_or_o[x][abs(y - 1)]
                    idx = [x, abs(y - 1)]

        if len(data.boards) == idx[0]:
            data.boards.append([" " for _ in range(9)])
        data.boards[idx[0]][move] = x_or_o[idx[0]][idx[1]]

        # Winning combos
        row1, row2, row3 = [data.boards[idx[0]][i:i + 3] for i in range(0, 9, 3)]
        col1, col2, col3 = [data.boards[idx[0]][i:7 + i:3] for i in range(3)]
        diag1, diag2 = [
            [data.boards[idx[0]][0], data.boards[idx[0]][4], data.boards[idx[0]][8]],
            [data.boards[idx[0]][2], data.boards[idx[0]][4], data.boards[idx[0]][6]]
        ]
        no_winner = True

        for check in (row1, row2, row3, col1, col2, col3, diag1, diag2):
            if (check[0] == check[1] == check[2]) and statistics.mode(check) != ' ':  # WINNER!
                no_winner = False
                winner = statistics.mode(check)  # Checks the most occurances
                winner = "X" if winner == "O" else "O"

                if winner == opponent_letter:
                    s.send_client(opponent_ip, "win", str(move).encode())
                    s.send_client(clt_data['ip'], 'lose', str(move).encode())
                else:
                    s.send_client(clt_data['ip'], "win", str(move).encode())
                    s.send_client(opponent_ip, 'lose', str(move).encode())

        spaces = [
            i == " " for i in data.boards[idx[0]]
        ]

        if not any(spaces):
            # No spaces and no winner - Tied
            no_winner = False  # Even though it's technically true
            s.send_client(opponent_ip, "tie", str(move).encode())
            s.send_client(clt_data['ip'], 'tie', str(move).encode())

        if no_winner:
            s.send_client(opponent_ip, "player_turn", str(move).encode())

    while True:
        try:
            # Runs server
            s.run()
        except utils.NoMessageException:
            pass


if __name__ == "__main__":
    run()



import sys
import pathlib
import pygame

from functools import lru_cache

PATH = pathlib.Path(__file__)

# Import stuff, as this situation's a bit wonky
# This is not needed in usual hisock code. However, in this odd case (two same named directories),
# imports need to be configured first like this
sys.path.append(str(PATH.parent.parent.parent))  # Not needed usually; used to run examples

from hisock.client import (
    threaded_connect
)


# Data class, as to prevent using those dirty globals
class Data:
    def __init__(self):
        self.tictactoe_opponent = None  # Tic Tac Toe opponent name
        self.board = [' ' for _ in range(9)]  # Tic Tac Toe board
        self.letter = ''
        self.turn = None
        self.game_start = False


@lru_cache
def load_font(size):
    return pygame.font.Font(None, size)


def run():
    pygame.init()

    ip_input = input("Enter IP of where server is hosted: ")

    port_input = input("Enter port of server: ")
    while port_input == '':
        port_input = input("Enter port of server: ")
    port_input = int(port_input)

    name = input("Enter name (MAX LENGTH: 32): ")
    while name == '' or len(name) > 32:
        name = input("Enter name (MAX LENGTH: 32): ")

    print(f"Connecting to server at {ip_input}:{port_input}...", end=" ")

    server = threaded_connect((ip_input, port_input), name)
    print("SUCCESS!")

    data = Data()

    @server.on("game_start")
    def game_start(opponent: str):
        data.game_start = True
        data.tictactoe_opponent = opponent
        goes_first = server.recv_raw()

        if goes_first == b"First":
            data.turn = True
            data.letter = "X"
        else:
            data.turn = False
            data.letter = "O"

    screen = pygame.display.set_mode((400, 400))

    server.start_client()

    opp_font = load_font(20)

    while True:
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                server.stop_client()
                pygame.quit()
                sys.exit()
            elif (
                    event.type == pygame.MOUSEBUTTONDOWN and
                    data.game_start and
                    data.turn
            ):
                bod_idx = [
                    xy // 133 for xy in event.pos
                ]

                board_idx = bod_idx[0] * 3 + bod_idx[1]
                data.board[board_idx] = data.letter
                print(bod_idx)

            screen.fill((51, 168, 12))
            pygame.draw.line(
                screen, (0, 0, 0),
                (133, 20), (133, 380),
                width=10
            )
            pygame.draw.line(
                screen, (0, 0, 0),
                (266, 20), (266, 380),
                width=10
            )
            pygame.draw.line(
                screen, (0, 0, 0),
                (20, 133), (380, 133),
                width=10
            )
            pygame.draw.line(
                screen, (0, 0, 0),
                (20, 266), (380, 266),
                width=10
            )
            
            if data.tictactoe_opponent is None:
                opp_txt = opp_font.render("Waiting for opponent...", True, (0, 0, 0))
            else:
                opp_txt = opp_font.render(f"Opponent: {data.tictactoe_opponent}", True, (0, 0, 0))

            name_txt = opp_font.render(f"You are: {name}", True, (0, 0, 0))

            if data.turn:
                turn_txt = opp_font.render("Your turn", True, (0, 0, 0))
            elif data.turn is not None:
                turn_txt = opp_font.render("Opponent's Turn", True, (0, 0, 0))
            else:
                turn_txt = None

            screen.blit(opp_txt, (0, 0))
            screen.blit(name_txt, (0, 20))

            if turn_txt is not None:
                screen.blit(turn_txt, opp_txt.get_rect(topright=(380, 0)))

            pygame.display.update()


if __name__ == "__main__":
    run()



from . import (
    tictactoe_client,
    tictactoe_server
)


def run(client_or_serv="input"):
    if client_or_serv == "input":
        user_input = input("Are you connecting to or hosting a tictactoe match? (C/H) ")

        if user_input.lower() == "c":
            tictactoe_client.run()
        elif user_input.lower() == "h":
            tictactoe_server.run()
        else:
            print("No option selected, aborting...")
    elif client_or_serv.lower() == "client":
        tictactoe_client.run()
    elif client_or_serv.lower() == "server":
        tictactoe_server.run()



"""
Basic example of the structure of `hisock`. This is the server script.
Not an advanced example, but gets the main advantages of hisock across
"""

# Builtin modules
import sys
import time
import random

from hisock import start_server, get_local_ip, iptup_to_str


def run():
    ADDR = get_local_ip()
    PORT = 6969

    if len(sys.argv) == 2:
        ADDR = sys.argv[1]
    if len(sys.argv) == 3:
        PORT = int(sys.argv[2])

    print(f"Serving at {ADDR}")
    server = start_server((ADDR, PORT))  # Starts the server

    # `join` is a reserved function - It will activate
    # whenever a new client joins
    @server.on("join")
    def client_join(client_data):
        # The `join` reserved function returns one argument;
        # The client data, with IP, name, and group
        # (Names and Groups are just an easy way to identify clients without IP)
        print(f"Cool, {':'.join(map(str, client_data['ip']))} joined!")
        if client_data['name'] is not None:
            print(f"    - With a sick name \"{client_data['name']}\", very cool!")
        if client_data['group'] is not None:
            print(f"    - In a sick group \"{client_data['group']}\", cool!")

        print("I'm gonna send them a quick hello message")

        # The current way to send; it takes in a tuple of IP,
        # The command to send (essentially what to label the message as),
        # And the content (Content must be in bytes for now, str support may be added)
        server.send_client(client_data['ip'], "hello_message", str(time.time()).encode())

    # Client SHOULD send a command `processing1`, soon after the server message
    # This will activate whenever a client sends a `processing1` command
    @server.on("processing1")
    def process(client, process_request: str):
        # Nonreserved functions take two arguments; the client data,
        # And the actual content
        # In HiSock, you are able to type cast the initially `bytes` content
        # Into another supported type, like str, int, or float

        # `iptup_to_str` converts a tuple IP into a string IP ('1.1.1.1', 4) -> 1.1.1.1:4
        print(f"\nAlright, looks like {iptup_to_str(client['ip'])} received the hello message, "
              "\nas now they're trying to compute something on the server, because they have "
              "potato computers")
        print("Their processing request is:", process_request)

        for _ in range(process_request.count("randnum")):
            randnum = str(random.randint(1, 100000000))
            process_request = process_request.replace("randnum", randnum, 1)

        result = eval(process_request)  # Insecure, but I'm lazy, so...
        print(f"Cool! The result is {result}! I'mma send it to the client")

        # Whenever you want to send multiple content rounds without
        # creating 500 commands, use `send_client_raw`. Along with
        # `recv_raw`, it can be used to send multiple content rounds.
        # Here, we are sending the result back to the client
        server.send_client_raw(client['ip'], str(result).encode())

    while True:
        server.run()


if __name__ == "__main__":
    run()



"""
Basic example of the structure of `hisock`. This is the client script.
Not an advanced example, but gets the main advantages of hisock across
"""

# Builtin module
import time

from hisock import connect, get_local_ip


def run():
    # Parse `input` results
    server_to_connect = input("Enter server IP to connect to (Press enter for default of your local IP): ")
    port = input("Enter Server port number (Press enter for default of 6969): ")

    if server_to_connect == '':
        server_to_connect = get_local_ip()

    if port == '':
        port = 6969
    else:
        port = int(port)

    name = input("Name? (Press enter for no name) ")
    group = input("Group? (Press enter for no group) ")

    print("======================================= ESTABLISHING CONNECTION =======================================")

    if name == '':
        name = None
    if group == '':
        group = None

    join_time = time.time()

    # Establishes connection to the server
    s = connect(
        (server_to_connect, port),
        name=name, group=group
    )

    @s.on("hello_message")
    def handle_hello(msg):
        print("Thanks, server, for sending a hello, just for me!")
        print(f"Looks like, the message was sent on timestamp {msg.decode()}, "
              f"which is just {round(float(msg.decode()) - join_time, 6) * 1000} milliseconds since the connection!")
        print("In response, I'm going to send the server a request to do some processing")

        s.send("processing1", b"randnum**2")
        result = int(s.recv_raw())

        print(f"WHOAAA! The result is {result}! Thanks server!")

    while True:
        s.update()


if __name__ == "__main__":
    run()



import pathlib
import os
import re
from setuptools import setup

ROOT = pathlib.Path(__file__).parent  # pathlib.Path object of root (/hisock/)

README = (ROOT / "README.md").read_text()  # Reads README
NAME = "hisock"

CONSTANT_TXT = (ROOT / "hisock/constants.py").read_text()

VERSION = re.search("__version__ = \"[^\n\"]+\"", CONSTANT_TXT).group()
VERSION = VERSION.lstrip("__version__ = \"").rstrip("\"")


class ANSIColors:
    PINK = '\033[95m'
    PURPLE = '\033[35m'
    BLUE = '\033[94m'
    CYAN = '\033[96m'
    GREEN = '\033[92m'
    DARK_GREEN = '\033[32m'
    YELLOW = '\033[93m'
    DARK_YELLOW = '\033[33m'
    RED = '\033[91m'
    DARK_RED = '\033[31m'
    END = '\033[0m'

    FONTCHANGE_BOLD = '\033[1m'
    FONTCHANGE_UNDERLINE = '\033[4m'


def color_text(txt, color=None, font_changes=None):
    """I copied this from... my own code -.-"""
    return f"""{f'{getattr(ANSIColors, color.upper().replace(" ", "_"))}' if color is not None else ''}{f'{getattr(ANSIColors, f"FONTCHANGE_{font_changes.upper()}")}' if font_changes is not None else ''}{txt}{f'{ANSIColors.END * sum([bool(i) for i in [color, font_changes]])}'}"""


print(color_text(f"Building version {VERSION}...", "green"))

raw_ex_subdirs = [
    content.relative_to(ROOT) for content in (ROOT / "examples").iterdir()
]
stringified_ex_subdirs = list(
    map(
        str, list(filter(pathlib.Path.is_dir, raw_ex_subdirs))
    )
)
ex_subdirs = [
    subdir.replace('\\', '.').replace('/', '.') for subdir in stringified_ex_subdirs
]

print_ex_subdirs = "- " + "\n- ".join(ex_subdirs)
print(
    color_text("[INFO] Collected the following examples modules:\n"
               f"{print_ex_subdirs}", "green")
)

setup(
    name=NAME,
    version=VERSION,
    description="A higher-level extension of the socket module, with simpler and more efficient usages",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/SSS-Says-Snek/hisock",
    author="SSS-Says-Snek",
    author_email="bmcomis2018@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9"
    ],
    install_requires=[
        "pytest>=6.2.5",
        "pycryptodome>=3.11"
    ],
    packages=[
        'hisock',
        'examples',
    ] + ex_subdirs,
    python_requires=">=3.7"
)

if os.path.exists("dist"):
    for i in os.listdir("dist"):
        if not (
                re.match(f"{NAME}-{VERSION}\.tar\.gz", i) or
                re.match(f"{NAME}-{VERSION}-py3-none-any\.whl", i)
        ):
            print(color_text(f"Removing old version {i}...", "red"))
            os.remove(os.path.join("dist", i))
